from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, Type

import polars as pl

from ...core import (
    AssetDataFailedToPersist,
    AssetDataFailedToRetrieve,
    DataPersistableProtocol,
    DataPersister,
)
from ..resources.databases import DatabaseResource
from ..resources.folder import FolderResource

if TYPE_CHECKING:
    from polars._typing import SchemaDict  # type:ignore


PolarsPersistable = DataPersistableProtocol[pl.DataFrame]


@dataclass
class PolarsPersisterFolderResource:
    folder: FolderResource
    format: Literal["csv", "parquet", "json", "ndjson"] = "ndjson"

    def save(self, asset: PolarsPersistable) -> None:
        path = asset.asset_id().as_path(suffix=self.format)

        match self.format:
            case "csv":
                with self.folder.open(path, "w") as f:
                    asset.data.write_csv(f, separator=";")

            case "parquet":
                with self.folder.open(path, "wb") as f:
                    asset.data.write_parquet(f)

            case "json":
                with self.folder.open(path, "w") as f:
                    asset.data.write_json(f)

            case "ndjson":
                with self.folder.open(path, "w") as f:
                    asset.data.write_ndjson(f)

    def load(self, asset: PolarsPersistable) -> None:
        path = asset.asset_id().as_path(suffix=self.format)
        schema: SchemaDict = asset.schema if hasattr(asset, "schema") else None  # type: ignore
        columns = list(schema.keys()) if schema else None  # type: ignore

        match self.format:
            case "csv":
                with self.folder.open(path, "r") as f:
                    asset.data = pl.read_csv(
                        f, separator=";", schema=schema, truncate_ragged_lines=True
                    )

            case "parquet":
                with self.folder.open(path, "rb") as f:
                    asset.data = pl.read_parquet(
                        f, schema=schema, allow_missing_columns=True, columns=columns
                    )

            case "json":
                with self.folder.open(path, "r") as f:
                    asset.data = pl.read_json(f, schema=schema)

            case "ndjson":
                with self.folder.open(path, "r") as f:
                    asset.data = pl.read_ndjson(f, schema=schema)


@dataclass
class PolarsPersisterDBResource:
    db: DatabaseResource
    if_table_exists: Literal["append", "replace", "fail"] = "replace"
    batch: int | None = None

    def save(self, asset: PolarsPersistable) -> None:
        name = asset.asset_id().as_path().name
        chunks = [asset.data]
        if self.batch:
            chunks = [
                asset.data[i : i + self.batch]
                for i in range(0, asset.data.height, self.batch)
            ]

        with self.db.get_connection() as con:
            first_batch = True
            for chunk in chunks:
                if first_batch:
                    chunk.write_database(
                        name, con, if_table_exists=self.if_table_exists
                    )
                    first_batch = False
                else:
                    chunk.write_database(name, con, if_table_exists="append")

        with self.db.get_connection() as con:
            asset.data.write_database(name, con, if_table_exists=self.if_table_exists)

    def load(self, asset: PolarsPersistable) -> None:
        name = asset.asset_id().as_path().name
        schema: SchemaDict = asset.schema if hasattr(asset, "schema") else None  # type: ignore
        columns = list(schema.keys()) if schema else None  # type: ignore

        query = f'select * from "{name}";'
        with self.db.get_connection() as con:
            data = pl.read_database(query, con, schema_overrides=schema)

        if columns:
            for c in data.columns:
                if c not in columns:
                    data = data.drop(c)

            for c in columns:
                if c not in data.columns:
                    data = data.with_columns(pl.lit(None).alias(c))

        asset.data = data


PolarsPersisterResource = PolarsPersisterFolderResource | PolarsPersisterDBResource


class PolarsPersister(DataPersister):
    """
    The InMemoryDataPersister provides functionality to register, load, save,
    and directly set data for assets, using an in-memory dictionary to store the data.
    """

    def __init__(self, resource: PolarsPersisterResource) -> None:
        self.resource = resource

    def register(
        self, *asset: PolarsPersistable | Type[PolarsPersistable], **kwargs: Any
    ) -> None:
        for a in asset:
            self.patch_asset(a)

    def save(self, asset: PolarsPersistable) -> None:
        try:
            self.resource.save(asset)
        except Exception as e:
            raise AssetDataFailedToPersist(asset.asset_id(), e) from e

    def load(self, asset: PolarsPersistable) -> None:
        try:
            self.resource.load(asset)
        except Exception as e:
            raise AssetDataFailedToRetrieve(asset.asset_id(), e) from e



from typing import Any

import pytest

from sidas.core import (
    AssetId,
    AssetNotRegisteredInDataPersister,
    AssetNotRegisteredInMetaPersister,
    AssetStatus,
    DefaultAsset,
)
from sidas.extensions.assets import SimpleAsset, SimpleAssetMetaData
from sidas.extensions.data_persisters import InMemoryDataPersister
from sidas.extensions.meta_persisters import InMemoryMetaPersister

# -----------------------------------------------------------------------------
# CONFIG
# -----------------------------------------------------------------------------
META_TYPE = SimpleAssetMetaData
DATA_TYPE = int
ASSED_IDENTIFIER = AssetId("myname")
TRANSFORMATION_RESULT = 10


class ExampleAsset(SimpleAsset[DATA_TYPE]):
    def transformation(self, *args: Any, **kwargs: Any) -> int:
        return TRANSFORMATION_RESULT


class ExampleAssetNamed(SimpleAsset[DATA_TYPE]):
    asset_identifier = ASSED_IDENTIFIER

    def transformation(self, *args: Any, **kwargs: Any) -> int:
        return TRANSFORMATION_RESULT


ASSET_CLASSES: list[type[DefaultAsset]] = [
    ExampleAsset,
    ExampleAssetNamed,
]


# -----------------------------------------------------------------------------
# Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture(autouse=True)
def clear_patched_methods():
    """Clear any patched load/save methods before and after each test to ensure clean state."""
    backup: dict[type[DefaultAsset], dict[str, Any]] = {
        asset_class: {} for asset_class in ASSET_CLASSES
    }
    for asset_class in ASSET_CLASSES:
        backup[asset_class]["load_data"] = asset_class.load_data
        backup[asset_class]["save_data"] = asset_class.save_data
        backup[asset_class]["save_meta"] = asset_class.save_meta
        backup[asset_class]["load_meta"] = asset_class.load_meta

    yield

    for asset_class in ASSET_CLASSES:
        asset_class.load_data = backup[asset_class]["load_data"]
        asset_class.save_data = backup[asset_class]["save_data"]
        asset_class.save_meta = backup[asset_class]["save_meta"]
        asset_class.load_meta = backup[asset_class]["load_meta"]


@pytest.fixture()
def data_persister() -> InMemoryDataPersister:
    return InMemoryDataPersister()


@pytest.fixture()
def meta_persister() -> InMemoryMetaPersister:
    return InMemoryMetaPersister()


@pytest.fixture()
def asset(
    data_persister: InMemoryDataPersister, meta_persister: InMemoryMetaPersister
) -> ExampleAsset:
    asset = ExampleAsset()
    data_persister.register(asset)
    meta_persister.register(asset)
    asset.initialize()
    return asset


# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------


# Class Methods
# -----------------------------------------------------------------------------
def test_asset_id_default():
    assert ExampleAsset.asset_id() == AssetId(
        "tests.test_assets.test_simple_asset.ExampleAsset"
    )


def test_asset_id_explicit():
    assert ExampleAssetNamed.asset_id() == ASSED_IDENTIFIER


def test_meta_type():
    assert ExampleAsset.meta_type() == META_TYPE


def test_data_type():
    assert ExampleAsset.data_type() == DATA_TYPE


# Methods
# -----------------------------------------------------------------------------
def test_init():
    asset = ExampleAsset()
    assert ExampleAsset.assets[ExampleAsset.asset_id()] == asset


def test_initialize_failes_when_not_registered_in_meta_data_persister(
    data_persister: InMemoryDataPersister,
):
    asset = ExampleAsset()
    data_persister.register(asset)
    with pytest.raises(AssetNotRegisteredInMetaPersister):
        asset.initialize()


def test_initialize_failes_when_not_registered_in_data_data_persister(
    meta_persister: InMemoryMetaPersister,
):
    asset = ExampleAsset()
    meta_persister.register(asset)
    with pytest.raises(AssetNotRegisteredInDataPersister):
        asset.initialize()

    assert meta_persister.meta(asset).status == AssetStatus.INITIALIZING_FAILED


def test_initialize_successfull(
    meta_persister: InMemoryMetaPersister,
    data_persister: InMemoryDataPersister,
):
    asset = ExampleAsset()
    meta_persister.register(asset)
    data_persister.register(asset)

    asset.initialize()
    assert meta_persister.meta(asset).status == AssetStatus.INITIALIZED


def test_transformation(asset: ExampleAsset):
    assert asset.transformation() == TRANSFORMATION_RESULT


def test_execute_transformation(asset: ExampleAsset):
    assert asset.execute_transformation() == TRANSFORMATION_RESULT


def test_can_materialize(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.INITIALIZING)
    asset.save_meta()
    assert not asset.can_materialize()

    asset.meta.update_status(AssetStatus.INITIALIZING_FAILED)
    asset.save_meta()
    assert not asset.can_materialize()

    asset.meta.update_status(AssetStatus.INITIALIZED)
    asset.save_meta()
    assert asset.can_materialize()

    asset.meta.update_status(AssetStatus.MATERIALIZING)
    asset.save_meta()
    assert not asset.can_materialize()

    asset.meta.update_status(AssetStatus.MATERIALIZING_FAILED)
    asset.save_meta()
    assert asset.can_materialize()

    asset.meta.update_status(AssetStatus.MATERIALIZED)
    asset.save_meta()
    assert not asset.can_materialize()


def test_materialize_success(
    asset: ExampleAsset,
    meta_persister: InMemoryMetaPersister,
    data_persister: InMemoryDataPersister,
):
    asset.initialize()
    asset.materialize()

    assert meta_persister.meta(asset).status == AssetStatus.MATERIALIZED
    assert data_persister.data[asset.asset_id()] == TRANSFORMATION_RESULT


def test_materialize_failed_transformation(
    asset: ExampleAsset,
    meta_persister: InMemoryMetaPersister,
):
    asset.transformation = None  # type: ignore
    asset.materialize()

    assert meta_persister.meta(asset).status == AssetStatus.MATERIALIZING_FAILED


def test_materialize_failed_persisting(
    asset: ExampleAsset,
    meta_persister: InMemoryMetaPersister,
):
    asset.save_data = None  # type: ignore
    asset.materialize()
    assert meta_persister.meta(asset).status == AssetStatus.MATERIALIZING_FAILED



from datetime import datetime
from typing import Any

import pytest
from freezegun import freeze_time

from sidas.core import (
    AssetId,
    AssetStatus,
    DefaultAsset,
)
from sidas.extensions.assets import (
    ScheduledAsset,
    ScheduledAssetMetadata,
)
from sidas.extensions.data_persisters import InMemoryDataPersister
from sidas.extensions.meta_persisters import InMemoryMetaPersister

# -----------------------------------------------------------------------------
# CONFIG
# -----------------------------------------------------------------------------
META_TYPE = ScheduledAssetMetadata
DATA_TYPE = int
CRON_EXPRESSION = "*/5 * * * *"
ASSED_IDENTIFIER = AssetId("myname")
TRANSFORMATION_RESULT = 10


class ExampleAsset(ScheduledAsset[DATA_TYPE]):
    cron_expression = CRON_EXPRESSION

    def transformation(self) -> int:
        return TRANSFORMATION_RESULT


ASSET_CLASSES: list[type[DefaultAsset]] = [ExampleAsset]


# -----------------------------------------------------------------------------
# Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture(autouse=True)
def clear_patched_methods():
    """Clear any patched load/save methods before and after each test to ensure clean state."""
    backup: dict[type[DefaultAsset], dict[str, Any]] = {
        asset_class: {} for asset_class in ASSET_CLASSES
    }
    for asset_class in ASSET_CLASSES:
        backup[asset_class]["load_data"] = asset_class.load_data
        backup[asset_class]["save_data"] = asset_class.save_data
        backup[asset_class]["save_meta"] = asset_class.save_meta
        backup[asset_class]["load_meta"] = asset_class.load_meta

    yield

    for asset_class in ASSET_CLASSES:
        asset_class.load_data = backup[asset_class]["load_data"]
        asset_class.save_data = backup[asset_class]["save_data"]
        asset_class.save_meta = backup[asset_class]["save_meta"]
        asset_class.load_meta = backup[asset_class]["load_meta"]


@pytest.fixture()
def data_persister() -> InMemoryDataPersister:
    return InMemoryDataPersister()


@pytest.fixture()
def meta_persister() -> InMemoryMetaPersister:
    return InMemoryMetaPersister()


@pytest.fixture()
def asset(
    data_persister: InMemoryDataPersister, meta_persister: InMemoryMetaPersister
) -> ExampleAsset:
    asset = ExampleAsset()
    data_persister.register(asset)
    meta_persister.register(asset)
    asset.initialize()
    return asset


# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------


# Class Methods
# -----------------------------------------------------------------------------
def test_meta_type():
    assert ExampleAsset.meta_type() == META_TYPE


def test_data_type():
    assert ExampleAsset.data_type() == DATA_TYPE


# Methods
# -----------------------------------------------------------------------------
@freeze_time("2023-01-01 10:00:00")
def test_set_default_meta():
    asset = ExampleAsset()
    meta = asset.set_default_meta()
    assert meta.cron_expression == CRON_EXPRESSION
    assert meta.next_schedule == datetime.now()


def test_can_materialize_initializing(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.INITIALIZING)
    asset.save_meta()
    assert not asset.can_materialize()


def test_can_materialize_initializing_failed(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.INITIALIZING_FAILED)
    asset.save_meta()
    assert not asset.can_materialize()


def test_can_materialize_initalized(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.INITIALIZED)
    asset.save_meta()
    assert asset.can_materialize()


def test_can_materialize_persisting(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.MATERIALIZING)
    asset.save_meta()
    assert not asset.can_materialize()


def test_can_materialize_persisting_failed(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.MATERIALIZING_FAILED)
    asset.save_meta()
    assert asset.can_materialize()


@freeze_time("2023-01-01 10:00:00")
def test_can_materialize_persisted_before_next(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.MATERIALIZED)
    asset.meta.next_schedule = datetime(2023, 1, 1, 10, 10)
    asset.save_meta()

    with freeze_time("2023-01-01 10:03:00"):
        assert not asset.can_materialize()


@freeze_time("2023-01-01 10:00:00")
def test_can_materialize_persisted_after_next(asset: ExampleAsset):
    asset.meta.update_status(AssetStatus.MATERIALIZED)
    asset.meta.next_schedule = datetime(2023, 1, 1, 10, 10)
    asset.save_meta()

    with freeze_time("2023-01-01 10:11:00"):
        assert asset.can_materialize()


def test_after_materialize(asset: ExampleAsset):
    asset.initialize()
    asset.materialize()



import types
from pathlib import Path
from typing import Any

import duckdb
import pandas as pd
import pytest
from polars._typing import SchemaDict  # type: ignore
from polars.datatypes import Int64, String

from sidas.extensions.assets import SimpleAsset
from sidas.extensions.data_persisters.duckdb_persister import (
    DuckDbPersister,
    DuckDbPersisterDBResource,
    DuckDbPersisterFolderResource,
)
from sidas.extensions.resources.databases import SqliteResource
from sidas.extensions.resources.folder import LocalFolder

NO_SCHEMA = None
MATCHING_SCHEMA: SchemaDict = {"a": Int64}
EXTENDING_SCHEMA: SchemaDict = {"a": Int64, "b": String}


@pytest.fixture()
def csv_persister(tmp_path: Path) -> DuckDbPersister:
    folder = LocalFolder(tmp_path)
    resource = DuckDbPersisterFolderResource(folder, "csv")
    persister = DuckDbPersister(resource)
    return persister


@pytest.fixture()
def parquet_persister(tmp_path: Path) -> DuckDbPersister:
    folder = LocalFolder(tmp_path)
    resource = DuckDbPersisterFolderResource(folder, "parquet")
    persister = DuckDbPersister(resource)
    return persister


@pytest.fixture()
def json_persister(tmp_path: Path) -> DuckDbPersister:
    folder = LocalFolder(tmp_path)
    resource = DuckDbPersisterFolderResource(folder, "json")
    persister = DuckDbPersister(resource)
    return persister


@pytest.fixture()
def ndjson_persister(tmp_path: Path) -> DuckDbPersister:
    folder = LocalFolder(tmp_path)
    resource = DuckDbPersisterFolderResource(folder, "ndjson")
    persister = DuckDbPersister(resource)
    return persister


@pytest.fixture()
def db_persister(tmp_path: Path) -> DuckDbPersister:
    file = SqliteResource(tmp_path / "test.db")
    resource = DuckDbPersisterDBResource(file)
    persister = DuckDbPersister(resource)
    return persister


@pytest.fixture()
def persisters(
    csv_persister: DuckDbPersister,
    parquet_persister: DuckDbPersister,
    json_persister: DuckDbPersister,
    ndjson_persister: DuckDbPersister,
    db_persister: DuckDbPersister,
) -> list[DuckDbPersister]:
    return [
        csv_persister,
        parquet_persister,
        json_persister,
        ndjson_persister,
        db_persister,
    ]


def example_asset(
    data: dict[str, list[Any]] | None, schema: None | SchemaDict = None
) -> SimpleAsset[duckdb.DuckDBPyRelation]:
    asset = types.new_class("ExampleAsset", [SimpleAsset[duckdb.DuckDBPyRelation]])
    asset.schema = schema
    if data:
        _ = pd.DataFrame(data)
        asset.data = duckdb.sql("SELECT * FROM _")
    return asset  # type: ignore


def test_persist_multi_overwrite(persisters: list[DuckDbPersister]):
    save_asset_1 = example_asset({"a": [10, 20]}, NO_SCHEMA)
    save_asset_2 = example_asset({"a": [30, 40]}, NO_SCHEMA)
    load_asset = example_asset({}, NO_SCHEMA)

    for persister in persisters:
        persister.save(save_asset_1)
        persister.save(save_asset_2)
        persister.load(load_asset)

        assert save_asset_2.data.df().equals(load_asset.data.df())  # type: ignore


def test_persist_no_schema(persisters: list[DuckDbPersister]):
    data = {"a": [10, 20]}

    save_asset = example_asset(data, NO_SCHEMA)
    load_asset = example_asset({}, NO_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)

        assert save_asset.data is not None
        assert load_asset.data is not None
        assert save_asset.data.df().equals(load_asset.data.df())  # type: ignore


def test_persist_matching_schema(
    persisters: list[DuckDbPersister],
):
    data = {"a": [10, 20]}

    save_asset = example_asset(data, MATCHING_SCHEMA)
    load_asset = example_asset({}, MATCHING_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)
        assert load_asset.data.df().to_dict() == pd.DataFrame(data=data).to_dict()


def test_persist_extending_schema(
    persisters: list[DuckDbPersister],
):
    data = {"a": [10, 20]}
    expected: dict[str, Any] = {**data, "b": [None, None]}

    save_asset = example_asset(data, EXTENDING_SCHEMA)
    load_asset = example_asset({}, EXTENDING_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)

        assert load_asset.data.df().to_dict() == pd.DataFrame(data=expected).to_dict()


def test_persist_reducing_schema(
    persisters: list[DuckDbPersister],
):
    data: dict[str, Any] = {"a": [10, 20], "b": ["x", "y"]}
    expected = {"a": [10, 20]}

    save_asset = example_asset(data, MATCHING_SCHEMA)
    load_asset = example_asset({}, MATCHING_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)

        assert load_asset.data.df().to_dict() == pd.DataFrame(data=expected).to_dict()



from datetime import datetime, timedelta

import pytest

from sidas.core import (
    AssetMetaData,
    AssetStatus,
)


def test_default_initialization():
    """Test that a new MetaBase instance has the expected default values."""
    before = datetime.now()
    meta = AssetMetaData()
    after = datetime.now()

    assert meta.status == AssetStatus.INITIALIZING

    assert meta.updated_at is not None
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.materializing_started_at is None
    assert meta.materializing_stopped_at is None

    assert before <= meta.updated_at <= after
    assert before <= meta.initializing_started_at <= after


def test_update_status_initialized():
    """Test updating status to INITIALIZED."""
    meta = AssetMetaData()
    meta.update_status(AssetStatus.INITIALIZED)

    assert meta.status == AssetStatus.INITIALIZED

    assert meta.updated_at is not None
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is not None
    assert meta.materializing_started_at is None
    assert meta.materializing_stopped_at is None

    assert meta.initializing_started_at <= meta.initializing_stopped_at


def test_update_status_materializing():
    """Test updating status to MATERIALIZING."""
    meta = AssetMetaData()
    meta.update_status(AssetStatus.MATERIALIZING)

    assert meta.status == AssetStatus.MATERIALIZING

    assert meta.updated_at is not None
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.materializing_started_at is not None
    assert meta.materializing_stopped_at is None

    assert meta.initializing_started_at <= meta.materializing_started_at


def test_update_status_materializing_failed():
    """Test updating status to MATERIALIZING_FAILED."""
    meta = AssetMetaData()
    meta.update_status(AssetStatus.MATERIALIZING_FAILED)

    assert meta.status == AssetStatus.MATERIALIZING_FAILED

    assert meta.updated_at is not None
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.materializing_started_at is None
    assert meta.materializing_stopped_at is not None

    assert meta.initializing_started_at <= meta.materializing_stopped_at


def test_update_status_materialized():
    """Test updating status to MATERIALIZED."""
    meta = AssetMetaData()
    meta.update_status(AssetStatus.MATERIALIZED)

    assert meta.status == AssetStatus.MATERIALIZED

    assert meta.updated_at is not None
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.materializing_started_at is None
    assert meta.materializing_stopped_at is not None

    assert meta.initializing_started_at <= meta.materializing_stopped_at


def test_status_chaining():
    """Test the full lifecycle of status updates with method chaining."""
    meta = AssetMetaData()

    # Test method chaining
    result = (
        meta.update_status(AssetStatus.INITIALIZED)
        .update_status(AssetStatus.MATERIALIZING)
        .update_status(AssetStatus.MATERIALIZED)
    )

    assert result is meta  # Confirm method chaining returns self
    assert meta.status == AssetStatus.MATERIALIZED

    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is not None
    assert meta.materializing_started_at is not None
    assert meta.materializing_stopped_at is not None

    assert meta.initializing_started_at <= meta.initializing_stopped_at
    assert meta.initializing_started_at <= meta.materializing_started_at
    assert meta.materializing_started_at <= meta.materializing_stopped_at


def test_blocked():
    """Test the in_progress method."""
    meta = AssetMetaData()

    meta.update_status(AssetStatus.INITIALIZING)
    assert meta.blocked()

    meta.update_status(AssetStatus.INITIALIZED)
    assert not meta.blocked()

    meta.update_status(AssetStatus.INITIALIZING_FAILED)
    assert meta.blocked()

    meta.update_status(AssetStatus.MATERIALIZING)
    assert meta.blocked()

    meta.update_status(AssetStatus.MATERIALIZING_FAILED)
    assert not meta.blocked()

    meta.update_status(AssetStatus.MATERIALIZED)
    assert not meta.blocked()


def test_has_error():
    """Test the has_error method."""
    meta = AssetMetaData()

    meta.update_status(AssetStatus.INITIALIZING)
    assert not meta.has_error()

    meta.update_status(AssetStatus.INITIALIZED)
    assert not meta.has_error()

    meta.update_status(AssetStatus.INITIALIZING_FAILED)
    assert not meta.has_error()

    meta.update_status(AssetStatus.MATERIALIZING)
    assert not meta.has_error()

    meta.update_status(AssetStatus.MATERIALIZING_FAILED)
    assert meta.has_error()

    meta.update_status(AssetStatus.MATERIALIZED)
    assert not meta.has_error()


def test_json_serialization():
    """Test serialization to and from JSON."""
    # Create a MetaBase with defined timestamps to avoid timing issues
    now = datetime.now()
    original = AssetMetaData(
        status=AssetStatus.MATERIALIZED,
        initializing_started_at=now - timedelta(minutes=10),
        initializing_stopped_at=now - timedelta(minutes=9),
        materializing_started_at=now - timedelta(minutes=3),
        materializing_stopped_at=now - timedelta(minutes=1),
        updated_at=now,
    )

    # Convert to JSON
    json_data = original.to_json()

    # Create a new instance from the JSON
    recreated = AssetMetaData.from_json(json_data)

    # Verify all fields match
    assert recreated.status == original.status
    assert recreated.initializing_started_at == original.initializing_started_at
    assert recreated.initializing_stopped_at == original.initializing_stopped_at
    assert recreated.materializing_started_at == original.materializing_started_at
    assert recreated.materializing_stopped_at == original.materializing_stopped_at
    assert recreated.updated_at == original.updated_at


def test_json_validation_error():
    """Test that invalid JSON data raises a validation error."""
    with pytest.raises(Exception):  # Pydantic will raise a validation error
        AssetMetaData.from_json('{"status": "INVALID_STATUS"}')



from .asset_id import AssetId


class CoordinatorNotRegisteredInMetaPersister(Exception):
    """
    Raised when a coordinator is not registered in the meta persister
    """

    def __init__(self):
        msg = "Coordinator is not registered in the meta persister."
        super().__init__(msg)


class AssetNotFoundException(Exception):
    """
    Raised when an asset is not found by the coordinator
    """

    def __init__(self, asset_id: AssetId):
        msg = f"Asset with id {asset_id} not found."
        super().__init__(msg)


class AssetNotRegisteredInMetaPersister(Exception):
    """
    Raised when an asset is not registered in the meta persister
    """

    def __init__(self, asset_id: AssetId):
        msg = f"Asset with id {asset_id} is not registered in the meta persister."
        super().__init__(msg)


class AssetNotRegisteredInDataPersister(Exception):
    """
    Raised when an asset is not registered in the data persister
    """

    def __init__(self, asset_id: AssetId):
        msg = f"Asset with id {asset_id} is not registered in the data persister."
        super().__init__(msg)


class AssetDataFailedToPersist(Exception):
    """
    Raised when a data persister is not able to persist the asset data
    """

    def __init__(self, asset_id: AssetId, error: Exception | str):
        msg = f"Failed to save data for asset {asset_id}: {error}"
        super().__init__(msg)


class AssetDataFailedToRetrieve(Exception):
    """
    Raised when a data persister is not able to retrieve the asset data
    """

    def __init__(self, asset_id: AssetId, error: Exception | str):
        msg = f"Failed to load data for asset {asset_id}: {error}"
        super().__init__(msg)


class MetaDataFailedToPersist(Exception):
    """
    Raised when a meta persister is not able to persist the asset metadata
    """

    def __init__(self, asset_id: AssetId, error: Exception | str):
        msg = f"Failed to persist meta data for asset {asset_id}: {error}"
        super().__init__(msg)


class MetaDataFailedToRetrieve(Exception):
    """
    Raised when a meta persister is not able to retrive the asset metadata
    """

    def __init__(self, asset_id: AssetId, error: Exception | str):
        msg = f"Failed to retrieve meta data for asset {asset_id}: {error}"
        super().__init__(msg)



from __future__ import annotations

import logging
import time
import traceback
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Type

from .asset import AssetId, DefaultAsset
from .exceptions import (
    AssetNotFoundException,
    CoordinatorNotRegisteredInMetaPersister,
    MetaDataFailedToRetrieve,
)
from .loader import load_assets
from .meta import AssetStatus, CoordinatorMetaData, CoordinatorStatus


class Coordinator(ABC):
    """
    A class responsible for managing and coordinating the data assets.
    The coordinator can start processing, load and save asset metadata, and materialize asset value.
    """

    @classmethod
    def asset_id(cls) -> AssetId:
        return AssetId("Coordinator")

    @classmethod
    def meta_type(cls) -> Type[CoordinatorMetaData]:
        return CoordinatorMetaData

    @staticmethod
    def load_coordinator() -> Coordinator:
        try:
            return load_assets(Coordinator)[0]
        except IndexError:
            raise Exception("Failed to load Coordinator Plugin")

    meta: CoordinatorMetaData

    def __init__(self, cron_expression: str | None = None) -> None:
        self.assets: list[DefaultAsset] = []
        self.cron_expression = cron_expression or "*/30 * * * * *"

    def register(self, *assets: DefaultAsset) -> None:
        for asset in assets:
            self.assets.append(asset)

    def load_meta(self) -> None:
        raise CoordinatorNotRegisteredInMetaPersister()

    def save_meta(self) -> None:
        raise CoordinatorNotRegisteredInMetaPersister()

    def initialize(self) -> CoordinatorStatus:
        try:
            self.load_meta()
        except MetaDataFailedToRetrieve:
            self.meta = CoordinatorMetaData(cron_expression=self.cron_expression)
            self.save_meta()

        try:
            for asset in self.assets:
                asset.initialize()
        except Exception as e:
            msg = f"Error validating assets: {e}"
            self.meta.update_status(CoordinatorStatus.INITIALIZING_FAILED)
            self.meta.update_log(msg)
            self.save_meta()
            logging.exception(msg)
            return CoordinatorStatus.INITIALIZING_FAILED

        self.meta.update_status(CoordinatorStatus.INITIALIZED)
        self.save_meta()
        return CoordinatorStatus.INITIALIZED

    def asset(self, asset_id: AssetId) -> DefaultAsset:
        for asset in self.assets:
            if asset.asset_id() == asset_id:
                return asset

        raise AssetNotFoundException(asset_id)

    @abstractmethod
    def trigger_materialization(self, asset: DefaultAsset) -> None:
        """
        Abstract method to kickoff the materialization of asset's value.
        This method should be implemented by subclasses.
        """

    def process(self, asset: DefaultAsset) -> None:
        logging.info("processing asset %s", asset.asset_id())

        # Check if we can materialize. If there is an error, log it and continue
        can_materialize = False
        try:
            can_materialize = asset.can_materialize()
        except Exception as e:
            msg = f"Exception in before_materialize: {str(e)}\n{traceback.format_exc()}"
            logging.exception(msg)
            self.meta.update_log(msg)
            self.save_meta()

        if not can_materialize:
            return

        # setting the assets status to MATERIALIZING
        asset.meta.update_status(AssetStatus.MATERIALIZING)
        asset.save_meta()

        # trigger materialization. If there is an error log it and set the asset status
        # to MATERIALIZING_FAILED
        try:
            self.trigger_materialization(asset)
        except Exception as e:
            asset.meta.update_status(AssetStatus.MATERIALIZING_FAILED)
            asset.save_meta()

            msg = f"Error processing assets: {e}"
            logging.exception(msg)
            self.meta.update_log(msg)
            self.save_meta()

    def run(self) -> None:
        status = self.initialize()
        if status != CoordinatorStatus.INITIALIZED:
            self.meta.terminate()

        while not self.meta.terminating():
            if datetime.now() < self.meta.next_schedule:
                time.sleep(10)
                self.load_meta()
                continue

            self.meta.update_status(CoordinatorStatus.PROCESSING)
            self.save_meta()

            for asset in self.assets:
                self.process(asset)

            self.meta.update_status(CoordinatorStatus.WAITING)
            self.save_meta()

        self.meta.update_status(CoordinatorStatus.TERMINATED)
        self.save_meta()
        return

    def materialize(self, asset_id: AssetId) -> None:
        asset = self.asset(asset_id)
        asset.load_meta()

        try:
            asset.before_materialize()
        except Exception as e:
            asset.meta.update_status(AssetStatus.MATERIALIZING_FAILED)
            asset.save_meta()

            msg = f"Exception in before_materialize: {str(e)}\n{traceback.format_exc()}"
            logging.exception(msg)
            self.meta.update_log(msg)
            return

        try:
            asset.materialize()
        except Exception as e:
            asset.meta.update_status(AssetStatus.MATERIALIZING_FAILED)
            asset.save_meta()

            msg = f"Exception in materialize: {str(e)}\n{traceback.format_exc()}"
            logging.exception(msg)
            self.meta.update_log(msg)
            return

        try:
            asset.after_materialize()
        except Exception as e:
            asset.meta.update_status(AssetStatus.MATERIALIZING_FAILED)
            asset.save_meta()

            msg = f"Exception in after_materialize: {str(e)}\n{traceback.format_exc()}"
            logging.exception(msg)
            self.meta.update_log(msg)



import types
from dataclasses import dataclass
from pathlib import Path
from typing import Type

import pytest
from polars._typing import SchemaDict
from polars.datatypes import Int32, String

from sidas.extensions.assets import SimpleAsset
from sidas.extensions.data_persisters.dataclass_persister import (
    DataclassPersister,
    DataclassPersisterDBResource,
    DataclassPersisterFolderResource,
)
from sidas.extensions.resources.databases import SqliteResource
from sidas.extensions.resources.folder import LocalFolder


@dataclass
class ExampleClass:
    x: int


NO_SCHEMA = None
MATCHING_SCHEMA: SchemaDict = {"a": Int32}
EXTENDING_SCHEMA: SchemaDict = {"a": Int32, "b": String}


@pytest.fixture()
def csv_persister(tmp_path: Path) -> DataclassPersister:
    folder = LocalFolder(tmp_path)
    resource = DataclassPersisterFolderResource(folder, "csv")
    persister = DataclassPersister(resource)
    return persister


@pytest.fixture()
def parquet_persister(tmp_path: Path) -> DataclassPersister:
    folder = LocalFolder(tmp_path)
    resource = DataclassPersisterFolderResource(folder, "parquet")
    persister = DataclassPersister(resource)
    return persister


@pytest.fixture()
def json_persister(tmp_path: Path) -> DataclassPersister:
    folder = LocalFolder(tmp_path)
    resource = DataclassPersisterFolderResource(folder, "json")
    persister = DataclassPersister(resource)
    return persister


@pytest.fixture()
def ndjson_persister(tmp_path: Path) -> DataclassPersister:
    folder = LocalFolder(tmp_path)
    resource = DataclassPersisterFolderResource(folder, "ndjson")
    persister = DataclassPersister(resource)
    return persister


@pytest.fixture()
def db_persister(tmp_path: Path) -> DataclassPersister:
    file = SqliteResource(tmp_path / "test.db")
    resource = DataclassPersisterDBResource(file)
    persister = DataclassPersister(resource)
    return persister


@pytest.fixture()
def persisters(
    csv_persister: DataclassPersister,
    parquet_persister: DataclassPersister,
    json_persister: DataclassPersister,
    ndjson_persister: DataclassPersister,
    db_persister: DataclassPersister,
) -> list[DataclassPersister]:
    return [
        csv_persister,
        parquet_persister,
        json_persister,
        ndjson_persister,
        db_persister,
    ]


def example_asset[T](
    signature: Type[T], data: T, schema: None | SchemaDict = None
) -> SimpleAsset[T]:
    asset = types.new_class("ExampleAsset", [SimpleAsset[signature]])
    asset.data = data
    asset.schema = schema
    return asset  # type: ignore


def test_persist_no_schema(persisters: list[DataclassPersister]):
    @dataclass
    class ExampleClass:
        a: int

    data = [ExampleClass(1), ExampleClass(2)]
    save_asset = example_asset(list[ExampleClass], data, NO_SCHEMA)
    load_asset = example_asset(list[ExampleClass], [], NO_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)
        assert save_asset.data == load_asset.data


def test_persist_matching_schema(
    persisters: list[DataclassPersister],
):
    @dataclass
    class ExampleClass:
        a: int

    data = [ExampleClass(1), ExampleClass(2)]
    save_asset = example_asset(list[ExampleClass], data, MATCHING_SCHEMA)
    load_asset = example_asset(list[ExampleClass], [], MATCHING_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)
        assert save_asset.data == load_asset.data


def test_persist_extending_schema(
    persisters: list[DataclassPersister],
):
    @dataclass
    class ExampleClass:
        a: int

    data = [ExampleClass(1), ExampleClass(2)]
    save_asset = example_asset(list[ExampleClass], data, EXTENDING_SCHEMA)
    load_asset = example_asset(list[ExampleClass], [], EXTENDING_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)
        assert save_asset.data == load_asset.data


def test_persist_reducing_schema(
    persisters: list[DataclassPersister],
):
    @dataclass
    class ExampleClass:
        a: int
        b: None | str

    data = [ExampleClass(1, "x"), ExampleClass(2, "x")]
    save_asset = example_asset(list[ExampleClass], data, MATCHING_SCHEMA)
    load_asset = example_asset(list[ExampleClass], [], MATCHING_SCHEMA)

    for persister in persisters:
        persister.save(save_asset)
        persister.load(load_asset)
        assert load_asset.data == [ExampleClass(1, None), ExampleClass(2, None)]



from datetime import datetime, timedelta

import pytest

from sidas.core import (
    AssetStatus,
    CoordinatorMetaData,
    CoordinatorStatus,
)

CRON_EXPRESSION = "*/30 * * * * *"


def test_default_initialization():
    """Test that a new MetaBase instance has the expected default values."""
    before = datetime.now()
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    after = datetime.now()

    assert meta.status == CoordinatorStatus.INITIALIZING
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is None

    assert before <= meta.updated_at <= after
    assert before <= meta.next_schedule <= after
    assert before <= meta.initializing_started_at <= after


def test_update_status_initialized():
    """Test updating status to INITIALIZED."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.INITIALIZED)

    assert meta.status == AssetStatus.INITIALIZED
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is not None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is None

    assert meta.initializing_started_at <= meta.updated_at
    assert meta.initializing_started_at <= meta.initializing_stopped_at


def test_update_status_hydrating():
    """Test updating status to INITIALIZING."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.INITIALIZING)

    assert meta.status == CoordinatorStatus.INITIALIZING
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is None

    assert meta.initializing_started_at <= meta.updated_at
    assert meta.initializing_started_at <= meta.initializing_started_at


def test_update_status_hydrating_failed():
    """Test updating status to INITIALIZING_FAILED."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.INITIALIZING_FAILED)

    assert meta.status == CoordinatorStatus.INITIALIZING_FAILED
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is not None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is None

    assert meta.initializing_started_at <= meta.updated_at
    assert meta.initializing_started_at <= meta.initializing_stopped_at


def test_update_status_hydrated():
    """Test updating status to INITIALIZED."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.INITIALIZED)

    assert meta.status == CoordinatorStatus.INITIALIZED
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is not None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is None

    assert meta.initializing_started_at <= meta.updated_at
    assert meta.initializing_started_at <= meta.initializing_stopped_at


def test_update_status_processing():
    """Test updating status to PROCESSING."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.PROCESSING)

    assert meta.status == CoordinatorStatus.PROCESSING
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.processing_started_at is not None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is None

    assert meta.initializing_started_at <= meta.updated_at
    assert meta.initializing_started_at <= meta.processing_started_at


def test_update_status_waiting():
    """Test updating status to WAITING."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.WAITING)

    assert meta.status == CoordinatorStatus.WAITING
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is not None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is None

    assert meta.initializing_started_at <= meta.updated_at


def test_update_status_terminating():
    """Test updating status to TERMINATING."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.TERMINATING)

    assert meta.status == CoordinatorStatus.TERMINATING
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is not None
    assert meta.terminating_stopped_at is None

    assert meta.initializing_started_at <= meta.updated_at
    assert meta.initializing_started_at <= meta.terminating_started_at


def test_update_status_terminated():
    """Test updating status to TERMINATED."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)
    meta.update_status(CoordinatorStatus.TERMINATED)

    assert meta.status == CoordinatorStatus.TERMINATED
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is None
    assert meta.processing_started_at is None
    assert meta.processing_stopped_at is None
    assert meta.terminating_started_at is None
    assert meta.terminating_stopped_at is not None

    assert meta.initializing_started_at <= meta.updated_at
    assert meta.initializing_started_at <= meta.terminating_stopped_at


def test_status_chaining():
    """Test the full lifecycle of status updates with method chaining."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)

    # Test method chaining
    result = (
        meta.update_status(CoordinatorStatus.INITIALIZED)
        .update_status(CoordinatorStatus.INITIALIZING)
        .update_status(CoordinatorStatus.INITIALIZED)
        .update_status(CoordinatorStatus.PROCESSING)
        .update_status(CoordinatorStatus.WAITING)
        .update_status(CoordinatorStatus.TERMINATING)
        .update_status(CoordinatorStatus.TERMINATED)
    )

    assert result is meta  # Confirm method chaining returns self
    assert meta.status == CoordinatorStatus.TERMINATED
    assert meta.initializing_started_at is not None
    assert meta.initializing_stopped_at is not None
    assert meta.processing_started_at is not None
    assert meta.processing_stopped_at is not None
    assert meta.terminating_started_at is not None
    assert meta.terminating_stopped_at is not None

    assert meta.initializing_stopped_at > meta.initializing_started_at
    assert meta.processing_stopped_at > meta.processing_started_at
    assert meta.terminating_started_at > meta.processing_stopped_at
    assert meta.terminating_stopped_at > meta.terminating_started_at


def test_in_progress():
    """Test the in_progress method."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)

    for value in CoordinatorStatus:
        meta.update_status(value)
        if value in (
            CoordinatorStatus.INITIALIZING,
            CoordinatorStatus.PROCESSING,
        ):
            assert meta.in_progress()
        else:
            assert not meta.in_progress()


def test_has_error():
    """Test the test_has_error method."""
    meta = CoordinatorMetaData(cron_expression=CRON_EXPRESSION)

    for value in CoordinatorStatus:
        meta.update_status(value)
        if value in (CoordinatorStatus.INITIALIZING_FAILED,):
            assert meta.has_error()
        else:
            assert not meta.has_error()


def test_json_serialization():
    """Test serialization to and from JSON."""
    # Create a MetaBase with defined timestamps to avoid timing issues
    now = datetime.now()
    original = CoordinatorMetaData(
        cron_expression="",
        status=CoordinatorStatus.TERMINATED,
        initializing_started_at=now - timedelta(minutes=10),
        initializing_stopped_at=now - timedelta(minutes=9),
        processing_started_at=now - timedelta(minutes=3),
        processing_stopped_at=now - timedelta(minutes=2),
        terminating_started_at=now - timedelta(minutes=1),
        terminating_stopped_at=now,
        updated_at=now,
    )

    # Convert to JSON
    json_data = original.to_json()

    # Create a new instance from the JSON
    recreated = CoordinatorMetaData.from_json(json_data)

    # Verify all fields match
    assert recreated.status == original.status
    assert recreated.initializing_started_at == original.initializing_started_at
    assert recreated.initializing_stopped_at == original.initializing_stopped_at
    assert recreated.initializing_started_at == original.initializing_started_at
    assert recreated.initializing_stopped_at == original.initializing_stopped_at
    assert recreated.processing_started_at == original.processing_started_at
    assert recreated.processing_stopped_at == original.processing_stopped_at
    assert recreated.terminating_started_at == original.terminating_started_at
    assert recreated.terminating_stopped_at == original.terminating_stopped_at
    assert recreated.updated_at == original.updated_at


def test_json_validation_error():
    """Test that invalid JSON data raises a validation error."""
    with pytest.raises(Exception):  # Pydantic will raise a validation error
        CoordinatorMetaData.from_json('{"status": "INVALID_STATUS"}')



from .dataclass_persister import (
    DataclassPersister,
    DataclassPersisterDBResource,
    DataclassPersisterFolderResource,
)
from .duckdb_persister import (
    DuckDbPersister,
    DuckDbPersisterDBResource,
    DuckDbPersisterFolderResource,
)
from .memory_persister import InMemoryDataPersister
from .pandas_persister import (
    PandasPersister,
    PandasPersisterDBResource,
    PandasPersisterFolderResource,
)
from .polars_persister import (
    PolarsPersister,
    PolarsPersisterDBResource,
    PolarsPersisterFolderResource,
)

__all__ = [
    "DataclassPersister",
    "DataclassPersisterDBResource",
    "DataclassPersisterFolderResource",
    "DuckDbPersister",
    "DuckDbPersisterFolderResource",
    "DuckDbPersisterDBResource",
    "PandasPersister",
    "PandasPersisterDBResource",
    "PandasPersisterFolderResource",
    "PolarsPersister",
    "PolarsPersisterDBResource",
    "PolarsPersisterFolderResource",
    "InMemoryDataPersister",
]



import subprocess

from ...core.asset import DefaultAsset
from ...core.coordinator import Coordinator


class SimpleCoordinator(Coordinator):
    def __init__(self, cron_expression: str | None = None) -> None:
        super().__init__(cron_expression=cron_expression)

    def trigger_materialization(self, asset: DefaultAsset) -> None:
        self.materialize(asset.asset_id())


class SimpleThreadedCoordinator(Coordinator):
    def __init__(self, cron_expression: str | None = None) -> None:
        super().__init__(cron_expression=cron_expression)

    def trigger_materialization(self, asset: DefaultAsset) -> None:
        asset_id = asset.asset_id()
        subprocess.run(["sidas", "materialize", asset_id])



from .downstream_asset import (
    CumulatingDownstreamAsset,
    DownstreamAsset,
    DownstreamAssetMetadata,
    DownstreamAssetRefreshMethod,
)
from .scheduled_asset import (
    CumulatingScheduledAsset,
    ScheduledAsset,
    ScheduledAssetMetadata,
)
from .simple_asset import SimpleAsset, SimpleAssetMetaData

__all__ = [
    "SimpleAsset",
    "SimpleAssetMetaData",
    "DownstreamAssetMetadata",
    "DownstreamAssetRefreshMethod",
    "DownstreamAsset",
    "CumulatingDownstreamAsset",
    "ScheduledAssetMetadata",
    "ScheduledAsset",
    "CumulatingScheduledAsset",
]



from typing import Any

import pytest

from sidas.core import AssetStatus, DefaultAsset
from sidas.extensions.assets import (
    DownstreamAsset,
    DownstreamAssetMetadata,
    DownstreamAssetRefreshMethod,
    SimpleAsset,
)
from sidas.extensions.data_persisters import InMemoryDataPersister
from sidas.extensions.meta_persisters import InMemoryMetaPersister

# -----------------------------------------------------------------------------
# CONFIG
# -----------------------------------------------------------------------------
META_TYPE = DownstreamAssetMetadata
DATA_TYPE = int
TRANSFORMATION_RESULT_UPSTREAM = 10
TRANSFORMATION_RESULT_DOWNSTREAM = 30


class Upstream(SimpleAsset[DATA_TYPE]):
    def transformation(self) -> int:
        return TRANSFORMATION_RESULT_UPSTREAM


class DownstreamAll(DownstreamAsset[DATA_TYPE]):
    refresh_method = DownstreamAssetRefreshMethod.ALL_UPSTREAM_REFRESHED

    def transformation(self, upstream: Upstream) -> int:
        return upstream.data + 20


class DownstreamAny(DownstreamAsset[DATA_TYPE]):
    refresh_method = DownstreamAssetRefreshMethod.ANY_UPSTREAM_REFRESHED

    def transformation(self, upstream: Upstream) -> int:
        return upstream.data + 20


ASSET_CLASSES: list[type[DefaultAsset]] = [Upstream, DownstreamAll, DownstreamAny]


# -----------------------------------------------------------------------------
# Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture(autouse=True)
def clear_patched_methods():
    """Clear any patched load/save methods before and after each test to ensure clean state."""
    backup: dict[type[DefaultAsset], dict[str, Any]] = {
        asset_class: {} for asset_class in ASSET_CLASSES
    }
    for asset_class in ASSET_CLASSES:
        backup[asset_class]["load_data"] = asset_class.load_data
        backup[asset_class]["save_data"] = asset_class.save_data
        backup[asset_class]["save_meta"] = asset_class.save_meta
        backup[asset_class]["load_meta"] = asset_class.load_meta

    yield

    for asset_class in ASSET_CLASSES:
        asset_class.load_data = backup[asset_class]["load_data"]
        asset_class.save_data = backup[asset_class]["save_data"]
        asset_class.save_meta = backup[asset_class]["save_meta"]
        asset_class.load_meta = backup[asset_class]["load_meta"]


@pytest.fixture()
def data_persister() -> InMemoryDataPersister:
    return InMemoryDataPersister()


@pytest.fixture()
def meta_persister() -> InMemoryMetaPersister:
    return InMemoryMetaPersister()


@pytest.fixture()
def upstream(
    data_persister: InMemoryDataPersister, meta_persister: InMemoryMetaPersister
) -> Upstream:
    asset = Upstream()
    data_persister.register(asset)
    meta_persister.register(asset)
    asset.initialize()
    return asset


@pytest.fixture()
def downstream_all(
    data_persister: InMemoryDataPersister, meta_persister: InMemoryMetaPersister
) -> DownstreamAll:
    asset = DownstreamAll()
    data_persister.register(asset)
    meta_persister.register(asset)
    asset.initialize()
    return asset


@pytest.fixture()
def downstream_any(
    data_persister: InMemoryDataPersister, meta_persister: InMemoryMetaPersister
) -> DownstreamAny:
    asset = DownstreamAny()
    data_persister.register(asset)
    meta_persister.register(asset)
    asset.initialize()
    return asset


# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------


# Class Methods
# -----------------------------------------------------------------------------
def test_meta_type():
    assert DownstreamAll.meta_type() == META_TYPE


def test_data_type():
    assert DownstreamAll.data_type() == DATA_TYPE


# Methods
# -----------------------------------------------------------------------------
def test_upstream(upstream: Upstream, downstream_all: DownstreamAll):
    assert downstream_all.upstream() == [upstream]


def test_set_default_meta_downstream_all(
    upstream: Upstream, downstream_all: DownstreamAll
):
    meta = downstream_all.set_default_meta()
    assert meta.refresh_method == DownstreamAssetRefreshMethod.ALL_UPSTREAM_REFRESHED
    assert meta.upstream == [upstream.asset_id()]


def test_set_default_meta_downstream_any(
    upstream: Upstream, downstream_any: DownstreamAny
):
    meta = downstream_any.set_default_meta()
    assert meta.refresh_method == DownstreamAssetRefreshMethod.ANY_UPSTREAM_REFRESHED
    assert meta.upstream == [upstream.asset_id()]


def test_can_materialize_initializing(
    upstream: Upstream, downstream_all: DownstreamAll, downstream_any: DownstreamAny
):
    downstream_all.meta.update_status(AssetStatus.INITIALIZING)
    downstream_all.save_meta()
    assert not downstream_all.can_materialize()

    downstream_any.meta.update_status(AssetStatus.INITIALIZING)
    downstream_any.save_meta()
    assert not downstream_any.can_materialize()


def test_can_materialize_initializing_failed(
    upstream: Upstream, downstream_all: DownstreamAll, downstream_any: DownstreamAny
):
    downstream_all.meta.update_status(AssetStatus.INITIALIZING_FAILED)
    downstream_all.save_meta()
    assert not downstream_all.can_materialize()

    downstream_any.meta.update_status(AssetStatus.INITIALIZING_FAILED)
    downstream_any.save_meta()
    assert not downstream_any.can_materialize()


def test_can_materialize_initialized(
    upstream: Upstream, downstream_all: DownstreamAll, downstream_any: DownstreamAny
):
    downstream_all.meta.update_status(AssetStatus.INITIALIZED)
    downstream_all.save_meta()

    downstream_any.meta.update_status(AssetStatus.INITIALIZED)
    downstream_any.save_meta()

    # upstream is initializing
    upstream.meta.update_status(AssetStatus.INITIALIZING)
    upstream.save_meta()
    assert not downstream_all.can_materialize()
    assert not downstream_any.can_materialize()

    # upstream failed to initialize
    upstream.meta.update_status(AssetStatus.INITIALIZING_FAILED)
    upstream.save_meta()
    assert not downstream_all.can_materialize()
    assert not downstream_any.can_materialize()

    # upstream is initialized
    upstream.meta.update_status(AssetStatus.INITIALIZED)
    upstream.save_meta()
    assert not downstream_all.can_materialize()
    assert not downstream_any.can_materialize()

    # upstream is materializing
    upstream.meta.update_status(AssetStatus.MATERIALIZING)
    upstream.save_meta()
    assert not downstream_all.can_materialize()
    assert not downstream_any.can_materialize()

    # upstream failed to materialize
    upstream.meta.update_status(AssetStatus.MATERIALIZING_FAILED)
    upstream.save_meta()
    assert not downstream_all.can_materialize()
    assert not downstream_any.can_materialize()

    # upstream materialized and downstream has not materialized
    upstream.meta.update_status(AssetStatus.MATERIALIZED)
    upstream.save_meta()
    assert downstream_all.can_materialize()
    assert downstream_any.can_materialize()

    # upstream materialized and downstream has materialized later
    downstream_all.meta.update_status(AssetStatus.MATERIALIZED)
    downstream_all.save_meta()
    downstream_any.meta.update_status(AssetStatus.MATERIALIZED)
    downstream_any.save_meta()
    assert not downstream_all.can_materialize()
    assert not downstream_any.can_materialize()


def test_before_materialize(
    upstream: Upstream,
    downstream_all: DownstreamAll,
    data_persister: InMemoryDataPersister,
):
    data_persister.data[upstream.asset_id()] = TRANSFORMATION_RESULT_UPSTREAM

    assert not hasattr(upstream, "data")
    downstream_all.before_materialize()

    assert hasattr(upstream, "data")
    assert upstream.data == TRANSFORMATION_RESULT_UPSTREAM



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, Type, get_args

import polars as pl
from dacite import from_dict

from ...core import (
    AssetDataFailedToPersist,
    AssetDataFailedToRetrieve,
    DataPersistableProtocol,
    DataPersister,
)
from ..resources.databases import DatabaseResource
from ..resources.folder import FolderResource

if TYPE_CHECKING:
    from polars._typing import SchemaDict  # type:ignore

DataclassPersistable = DataPersistableProtocol[list[Any]]


@dataclass
class DataclassPersisterFolderResource:
    folder: FolderResource
    file_format: Literal["csv", "parquet", "json", "ndjson"] = "ndjson"
    strict: bool = False

    def save(self, asset: DataclassPersistable) -> None:
        path = asset.asset_id().as_path(suffix=self.file_format)
        schema = asset.schema if hasattr(asset, "schema") else None  # type: ignore
        data = pl.DataFrame(asset.data, schema=schema, strict=self.strict, orient="row")  # type: ignore

        match self.file_format:
            case "csv":
                with self.folder.open(path, "w") as f:
                    data.write_csv(f, separator=";")

            case "parquet":
                with self.folder.open(path, "wb") as f:
                    data.write_parquet(f)

            case "json":
                with self.folder.open(path, "w") as f:
                    data.write_json(f)

            case "ndjson":
                with self.folder.open(path, "w") as f:
                    data.write_ndjson(f)

    def load(self, asset: DataclassPersistable) -> None:
        path = asset.asset_id().as_path(suffix=self.file_format)
        schema: SchemaDict = asset.schema if hasattr(asset, "schema") else None  # type: ignore

        match self.file_format:
            case "csv":
                with self.folder.open(path, "r") as f:
                    data = pl.read_csv(f, separator=";", schema=schema)

            case "parquet":
                with self.folder.open(path, "rb") as f:
                    data = pl.read_parquet(f, schema=schema)

            case "json":
                with self.folder.open(path, "r") as f:
                    data = pl.read_json(f, schema=schema)

            case "ndjson":
                with self.folder.open(path, "r") as f:
                    data = pl.read_ndjson(f, schema=schema)

        asset_type = get_args(asset.data_type())[0]
        asset.data = [from_dict(data_class=asset_type, data=d) for d in data.to_dicts()]


@dataclass
class DataclassPersisterDBResource:
    db: DatabaseResource
    if_table_exists: Literal["append", "replace", "fail"] = "replace"
    strict: bool = False
    batch: int | None = None

    def save(self, asset: DataclassPersistable) -> None:
        name = asset.asset_id().as_path().name
        schema: SchemaDict = asset.schema if hasattr(asset, "schema") else None  # type: ignore
        data = pl.DataFrame(asset.data, schema=schema, strict=self.strict)

        chunks = [data]
        if self.batch:
            chunks = [
                data[i : i + self.batch] for i in range(0, data.height, self.batch)
            ]

        with self.db.get_connection() as con:
            first_batch = True
            for chunk in chunks:
                if first_batch:
                    chunk.write_database(
                        name, con, if_table_exists=self.if_table_exists
                    )
                    first_batch = False
                else:
                    chunk.write_database(name, con, if_table_exists="append")

    def load(self, asset: DataclassPersistable) -> None:
        name = asset.asset_id().as_path().name
        schema: SchemaDict = asset.schema if hasattr(asset, "schema") else None  # type: ignore
        query = f'select * from "{name}";'
        with self.db.get_connection() as con:
            data = pl.read_database(query, con, schema_overrides=schema)

        asset_type = get_args(asset.data_type())[0]
        asset.data = [from_dict(data_class=asset_type, data=d) for d in data.to_dicts()]


DataclassPersisterResource = (
    DataclassPersisterFolderResource | DataclassPersisterDBResource
)


class DataclassPersister(DataPersister):
    """
    The InMemoryDataPersister provides functionality to register, load, save,
    and directly set data for assets, using an in-memory dictionary to store the data.
    """

    def __init__(
        self, resource: DataclassPersisterResource, strict: bool = False
    ) -> None:
        self.resource = resource

    def register(
        self, *asset: DataclassPersistable | Type[DataclassPersistable], **kwargs: Any
    ) -> None:
        for a in asset:
            self.patch_asset(a)

    def save(self, asset: DataclassPersistable) -> None:
        try:
            self.resource.save(asset)
        except Exception as e:
            raise AssetDataFailedToPersist(asset.asset_id(), e) from e

    def load(self, asset: DataclassPersistable) -> None:
        try:
            self.resource.load(asset)
        except Exception as e:
            raise AssetDataFailedToRetrieve(asset.asset_id(), e) from e



from typing import Any, Type

from ...core import (
    AssetId,
    AssetMetaData,
    MetaDataFailedToPersist,
    MetaDataFailedToRetrieve,
    MetaPersistable,
    MetaPersister,
)


class InMemoryMetaPersister(MetaPersister):
    def __init__(self) -> None:
        self._data: dict[AssetId, str] = {}

    def register(
        self, *asset: MetaPersistable | Type[MetaPersistable], **kwargs: Any
    ) -> None:
        for a in asset:
            self.patch_asset(a)

    def save(self, asset: MetaPersistable) -> None:
        try:
            self._data[asset.asset_id()] = asset.meta.to_json()
        except Exception as e:
            raise MetaDataFailedToPersist(asset.asset_id(), e) from e

    def load(self, asset: MetaPersistable) -> None:
        try:
            meta_raw = self._data[asset.asset_id()]
            meta = asset.meta_type().from_json(meta_raw)
            asset.meta = meta
        except Exception as e:
            raise MetaDataFailedToRetrieve(asset.asset_id(), e) from e

    def heartbeat(self) -> None:
        pass

    def meta(self, asset: MetaPersistable) -> AssetMetaData:
        meta_raw = self._data[asset.asset_id()]
        return asset.meta_type().from_json(meta_raw)


__all__ = ["InMemoryMetaPersister"]



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, Type

import duckdb
import polars as pl

from ...core import (
    AssetDataFailedToPersist,
    AssetDataFailedToRetrieve,
    DataPersistableProtocol,
    DataPersister,
)
from ..resources.databases import DatabaseResource
from ..resources.folder import FolderResource

if TYPE_CHECKING:
    from polars._typing import SchemaDict  # type:ignore
DuckDbPersistable = DataPersistableProtocol[duckdb.DuckDBPyRelation]


@dataclass
class DuckDbPersisterFolderResource:
    folder: FolderResource
    format: Literal["csv", "parquet", "json", "ndjson"] = "ndjson"

    def save(self, asset: DuckDbPersistable) -> None:
        path = asset.asset_id().as_path(suffix=self.format)
        data = asset.data.pl()

        match self.format:
            case "csv":
                with self.folder.open(path, "w") as f:
                    data.write_csv(f, separator=";")

            case "parquet":
                with self.folder.open(path, "wb") as f:
                    data.write_parquet(f)

            case "json":
                with self.folder.open(path, "w") as f:
                    data.write_json(f)

            case "ndjson":
                with self.folder.open(path, "w") as f:
                    data.write_ndjson(f)

    def load(self, asset: DuckDbPersistable) -> None:
        path = asset.asset_id().as_path(suffix=self.format)
        name = asset.asset_id().as_path().name
        schema: SchemaDict = asset.schema if hasattr(asset, "schema") else None  # type: ignore
        columns = list(schema.keys()) if schema else None  # type: ignore

        match self.format:
            case "csv":
                with self.folder.open(path, "r") as f:
                    data = pl.read_csv(
                        f, separator=";", schema=schema, truncate_ragged_lines=True
                    )

            case "parquet":
                with self.folder.open(path, "rb") as f:
                    data = pl.read_parquet(
                        f, schema=schema, allow_missing_columns=True, columns=columns
                    )

            case "json":
                with self.folder.open(path, "r") as f:
                    data = pl.read_json(f, schema=schema)

            case "ndjson":
                with self.folder.open(path, "r") as f:
                    data = pl.read_ndjson(f, schema=schema)

        # register the data alias (only used to avoid unused variable hints)
        duckdb.register("__data__", data)

        # load the data into the asset.data
        asset.data = duckdb.sql("select * from __data__;")

        # load the dat into the duckdb context
        try:
            duckdb.sql(f"drop table {name};")
        except duckdb.CatalogException:
            pass
        duckdb.sql(f"create table {name} as select * from __data__;")


@dataclass
class DuckDbPersisterDBResource:
    db: DatabaseResource
    if_table_exists: Literal["append", "replace", "fail"] = "replace"

    def save(self, asset: DuckDbPersistable) -> None:
        name = asset.asset_id().as_path().name
        data = asset.data.pl()
        with self.db.get_connection() as con:
            data.write_database(name, con, if_table_exists=self.if_table_exists)

    def load(self, asset: DuckDbPersistable) -> None:
        name = asset.asset_id().as_path().name
        schema: SchemaDict = asset.schema if hasattr(asset, "schema") else None  # type: ignore
        columns = list(schema.keys()) if schema else None  # type: ignore
        query = f'select * from "{name}";'

        with self.db.get_connection() as con:
            data = pl.read_database(query, con, schema_overrides=schema)

        if columns:
            for c in data.columns:
                if c not in columns:
                    data = data.drop(c)

            for c in columns:
                if c not in data.columns:
                    data = data.with_columns(pl.lit(None).alias(c))

        # register the data alias (only used to avoid unused variable hints)
        duckdb.register("__data__", data)

        # load the data into the asset.data
        asset.data = duckdb.sql("select * from __data__;")

        # load the dat into the duckdb context
        try:
            duckdb.sql(f"drop table {name};")
        except duckdb.CatalogException:
            pass
        duckdb.sql(f"create table {name} as select * from __data__;")


DuckDbPersisterResource = DuckDbPersisterFolderResource | DuckDbPersisterDBResource


class DuckDbPersister(DataPersister):
    """
    The InMemoryDataPersister provides functionality to register, load, save,
    and directly set data for assets, using an in-memory dictionary to store the data.
    """

    def __init__(self, resource: DuckDbPersisterResource) -> None:
        self.resource = resource

    def register(
        self, *asset: DuckDbPersistable | Type[DuckDbPersistable], **kwargs: Any
    ) -> None:
        for a in asset:
            self.patch_asset(a)

    def save(self, asset: DuckDbPersistable) -> None:
        try:
            self.resource.save(asset)
        except Exception as e:
            raise AssetDataFailedToPersist(asset.asset_id(), e) from e

    def load(self, asset: DuckDbPersistable) -> None:
        try:
            self.resource.load(asset)
        except Exception as e:
            raise AssetDataFailedToRetrieve(asset.asset_id(), e) from e



from typing import Any

import pytest

from sidas.core import (
    AssetId,
    AssetNotFoundException,
    AssetStatus,
    CoordinatorMetaData,
    CoordinatorStatus,
    DefaultAsset,
)
from sidas.extensions.assets import SimpleAsset
from sidas.extensions.coordinators import SimpleCoordinator
from sidas.extensions.data_persisters import InMemoryDataPersister
from sidas.extensions.meta_persisters import InMemoryMetaPersister

# -----------------------------------------------------------------------------
# CONFIG
# -----------------------------------------------------------------------------
CRON_EXPRESSION = "*/10 * * * * *"
ASSET_DATA = 10


class ExampleAsset(SimpleAsset[int]):
    def __init__(self) -> None:
        super().__init__()
        self.before_materialize_executed = False
        self.after_materialize_executed = False

    def transformation(self, *args: Any, **kwargs: Any) -> int:
        return ASSET_DATA

    def before_materialize(self) -> None:
        self.before_materialize_executed = True
        return super().before_materialize()

    def after_materialize(self) -> None:
        self.after_materialize_executed = True
        return super().after_materialize()


ASSET_CLASSES: list[type[DefaultAsset]] = [ExampleAsset]


# -----------------------------------------------------------------------------
# Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture(autouse=True)
def clear_patched_methods():
    """Clear any patched load/save methods before and after each test to ensure clean state."""
    backup: dict[type[DefaultAsset], dict[str, Any]] = {
        asset_class: {} for asset_class in ASSET_CLASSES
    }
    for asset_class in ASSET_CLASSES:
        backup[asset_class]["load_data"] = asset_class.load_data
        backup[asset_class]["save_data"] = asset_class.save_data
        backup[asset_class]["save_meta"] = asset_class.save_meta
        backup[asset_class]["load_meta"] = asset_class.load_meta

    yield

    for asset_class in ASSET_CLASSES:
        asset_class.load_data = backup[asset_class]["load_data"]
        asset_class.save_data = backup[asset_class]["save_data"]
        asset_class.save_meta = backup[asset_class]["save_meta"]
        asset_class.load_meta = backup[asset_class]["load_meta"]


@pytest.fixture()
def data_persister() -> InMemoryDataPersister:
    return InMemoryDataPersister()


@pytest.fixture()
def meta_persister() -> InMemoryMetaPersister:
    return InMemoryMetaPersister()


@pytest.fixture()
def asset(
    data_persister: InMemoryDataPersister, meta_persister: InMemoryMetaPersister
) -> ExampleAsset:
    asset = ExampleAsset()
    data_persister.register(asset)
    meta_persister.register(asset)
    return asset


@pytest.fixture()
def coordinator(meta_persister: InMemoryMetaPersister) -> SimpleCoordinator:
    coordinator = SimpleCoordinator(cron_expression=CRON_EXPRESSION)
    meta_persister.register(coordinator)
    return coordinator


# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------


# Class Methods
# -----------------------------------------------------------------------------
def test_assed_id():
    assert SimpleCoordinator.asset_id() == AssetId("Coordinator")


def test_meta_type():
    assert SimpleCoordinator.meta_type() == CoordinatorMetaData


def test_load_coordinator():
    pass


# Class Methods
# -----------------------------------------------------------------------------
def test_init(coordinator: SimpleCoordinator) -> None:
    assert coordinator
    assert coordinator.assets == []
    assert coordinator.cron_expression == CRON_EXPRESSION


def test_register(coordinator: SimpleCoordinator, asset: ExampleAsset):
    coordinator.register(asset)
    assert coordinator.assets == [asset]


def test_initialize(coordinator: SimpleCoordinator) -> None:
    assert not hasattr(coordinator, "meta")

    coordinator.initialize()
    assert hasattr(coordinator, "meta")
    assert isinstance(coordinator.meta, coordinator.meta_type())
    assert coordinator.meta.status == CoordinatorStatus.INITIALIZED


def test_initialize_with_assets(
    coordinator: SimpleCoordinator, asset: ExampleAsset
) -> None:
    coordinator.register(asset)

    coordinator.initialize()
    assert hasattr(asset, "meta")
    assert isinstance(asset.meta, asset.meta_type())
    assert asset.meta.status == AssetStatus.INITIALIZED


def test_initialize_with_assets_fails_asset_invalid(
    coordinator: SimpleCoordinator,
) -> None:
    """
    Unregistered asset fails validation
    """
    asset = ExampleAsset()
    coordinator.register(asset)

    coordinator.initialize()
    assert coordinator.meta.status == AssetStatus.INITIALIZING_FAILED


def test_get_asset(coordinator: SimpleCoordinator, asset: ExampleAsset):
    with pytest.raises(AssetNotFoundException):
        coordinator.asset(asset.asset_id())

    coordinator.register(asset)

    assert coordinator.asset(asset.asset_id()) == asset


def test_materialize(
    meta_persister: InMemoryMetaPersister,
    data_persister: InMemoryDataPersister,
    coordinator: SimpleCoordinator,
    asset: ExampleAsset,
):
    coordinator.register(asset)
    coordinator.initialize()
    coordinator.materialize(asset.asset_id())

    assert asset.before_materialize_executed
    assert asset.after_materialize_executed
    assert data_persister.data[asset.asset_id()] == ASSET_DATA
    assert meta_persister.meta(asset).status == AssetStatus.MATERIALIZED


def test_trigger_materialization(
    meta_persister: InMemoryMetaPersister,
    data_persister: InMemoryDataPersister,
    coordinator: SimpleCoordinator,
    asset: ExampleAsset,
):
    coordinator.register(asset)
    coordinator.initialize()
    coordinator.trigger_materialization(asset)

    assert asset.before_materialize_executed
    assert asset.after_materialize_executed
    assert data_persister.data[asset.asset_id()] == ASSET_DATA
    assert meta_persister.meta(asset).status == AssetStatus.MATERIALIZED


def test_materialize_terminate():
    pass
